#include <SoftwareSerial.h> // Library for Bluetooth module
#include <SPI.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <Arduino.h>

/*
  Le seguenti #define hanno il solo scopo mnemonico, nel programma sono stati utilizzati
  i registri relativi alle uscite
  Una modifica di queste define non comporta una modifica dell'uscita o della sua funzione
*/
#define RXPIN    4  // HC-05 tx
#define TXPIN    5  // HC-05 rx
#define BCSEL    8  // TLC Grayscale* / Brightness Control
#define SCLK     13 // Clock for data out
#define SIN      11 // Arduino data out - TLC data in
#define XLAT     7  // Shift to first data latch
#define BLANK0   6  // Blank pin for digit 1
#define BLANK1   9  // Blank pin for digit 2
#define BLANK2   2  // Blank pin for digit 3

int i, t=0;
int lum1, lum2, lum3; // Brightness value for each digit
volatile byte disp = 0;
SoftwareSerial bluetooth(RXPIN, TXPIN); // Start bluetooth radio

void setup() {
  //pinMode(BCSEL, OUTPUT);
  //pinMode(SCLK, OUTPUT);
  //pinMode(SIN, OUTPUT);
  //pinMode(XLAT, OUTPUT);
  //pinMode(BLANK0, OUTPUT);digitalWrite(BLANK0, HIGH); // Set output pin off
  //pinMode(BLANK1, OUTPUT);digitalWrite(BLANK1, HIGH); // Set output pin off
  //pinMode(BLANK2, OUTPUT);digitalWrite(BLANK2, HIGH);

  //digitalWrite(XLAT, LOW);
  //digitalWrite(BCSEL, LOW);
  //digitalWrite(SCLK, LOW);

  DDRD  |= 0b11100100; // Pin 0 to 7
  PORTD |= 0b01000100; // Set pin BLANK2 and BLANK0 high
  
  DDRB  |= 0b00111111; // Pin 8 to 13 as output (6' & 7'bit crystal)
  PORTB |= 0b00000110; // Set pin BLANK1 high

  SPI.begin();           // Use SPI hardaware to send data to TLC
  bluetooth.begin(9600); // Start bluetooth

  cli();                    // Disable global interrupts
  TCCR2A = 0;
  TCCR2B = 0;

  //TCNT2 = 1024;             // preload timer 65536-16MHz/256/2Hz
  TCCR2B |= (1 << CS22) | (1 << CS20);    // 128 prescaler
  TIMSK2 |= (1 << TOIE2);   // enable timer overflow interrupt
  sei();                    // Enable global interrupts
  
// **********************************************************************
// *  Initialize brightness shift                                       *
// **********************************************************************
  setBrightness(0b11101111, 0b11101111, 0b11101111);

// **********************************************************************
// *  Write initial status, half bottom segment on, other segments off  *
// **********************************************************************
  digitalWrite(BCSEL, LOW);
  SPI.beginTransaction(SPISettings(14000000, MSBFIRST, SPI_MODE3)); // Set SPI option (data read on low-to-high etc)
  for(i=0; i<6; i++){         
    prOut(0x0000);
  }
  prOut(0x00FF);
  for(i=0; i<9; i++){
    prOut(0x0000);
  }
  for(i=0; i<6; i++){      
    prOut(0x0000);
  }
  prOut(0x00FF);
  for(i=0; i<9; i++){
    prOut(0x0000);
  }
  SPI.endTransaction();

  writeLatch(); // Toggle latch pin
}

ISR(TIMER2_OVF_vect){        // interrupt service routine
  //TCNT1 = 34286;            // preload timer
  PORTD |= 0b01000100; // Off BLANK0 & BLANK2
  PORTB |= 0b00000010; // Off BLANK1
  disp += 1;
  if(disp == 3) disp = 0;
  if(disp == 0) PORTD &= ~0x64; // Low BLANK0
  if(disp == 1) PORTB &= ~0x02; // Low BLANK1
  if(disp == 2) PORTD &= ~0x04; // Low BLANK2
}

void loop() {
  
  if (bluetooth.available()){
    char inc = bluetooth.read();
    char rec[10];
      
    if(inc == '^'){ // Data that begin with ^ are command data
      unsigned long currTime = millis();
      while(bluetooth.available() == 0 && ((millis() - currTime) < 200));
      if((millis() - currTime) >= 200){
        bluetooth.println("Tempo massimo superato!");
      } else {
        inc = bluetooth.read();
        if(inc == '\n' || inc == '\r'){
          bluetooth.println("Formato comandi errato!");
        } else {
          if(inc == 'L' || inc == 'l'){ // Lux command sent, set led current
            i=0;
            currTime = millis();
            while((inc != '\n' || inc != '\r') && (millis() - currTime) < 200){
              if(bluetooth.available()){
                inc = bluetooth.read();
                rec[i++] = inc;
              }
            }
            int lum = 0; // Reset lum value
            if(i-1 >= 0) lum  = rec[i-1] - '0'; // Convert from char to int, max 3 digit (0-999)
            if(i-2 >= 0) lum += (rec[i-2] - '0') * 10;
            if(i-3 >= 0) lum += (rec[i-3] - '0') * 100;
            if(lum >= 0 && lum <= 127){ // Value is accetable
              lum1 = lum2 = lum3 = lum;
              setBrightness(lum1, lum2, lum3);
            }else{
              bluetooth.println("Value must be 0 < x < 127");
            }
          }
        }
      }//Close else{
    } else { // If char received is not a command or anithing else...
      rec[0] = inc;
      i=0;
      unsigned long currTime = millis();
      while((inc != '\n' || inc != '\r') && (millis() - currTime) < 200 && i < 2){
        if(bluetooth.available()){
          inc = bluetooth.read();
          if(inc != '\n' && inc != '\r') rec[++i] = inc;
        }
      }
      
      SPI.beginTransaction(SPISettings(14000000, MSBFIRST, SPI_MODE3));
      int j = 0;
      for(; j < 3; j++, i--){
        if(i >= 0) writeDigit(rec[i]); // Write on TLC each digit received
        else writeDigit('/');          // Blank digit if not written
      }
      SPI.endTransaction();
      writeLatch();
      while(bluetooth.available()) _delay_ms(2UL); // Empty bluetooth received data
    }
    
  }
  
  delay(2);
}

void writeLatch(){
  _delay_ms(1UL);
  PORTD |= 0b10000000;  // XLAT high
  _delay_ms(1UL);
  PORTD &= 0b01111111;  // XLAT low
}

void writeDigit(char a){
  PORTB &= 0b11111110; // Select grayscale shift register (BCSEL low)
  switch(a){
    case '0':
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case '1':
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      break;
    case '2':
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case '3':
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case '4':
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      break;
    case '5':
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      break;
    case '6':
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      break;
    case '7':
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case '8':
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case '9':
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case 'B':
    case 'b':
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      break;
    case 'A':
    case 'a':
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      prOut(0xFFFF);
      break;
    case '-':
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0xFFFF);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      break;
    default:        // Digit off
    //case '/':
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
      prOut(0x0000);
  }
}

void prOut(uint16_t data){
  TLCtransfer(data>>8 & 0xFF);
  TLCtransfer(data & 0xFF);
}

void TLCtransfer(uint8_t data){
  SPDR = data;
  asm volatile("nop");
  while (!(SPSR & _BV(SPIF))) ; // wait
}

void setBrightness(uint8_t b1, uint8_t b2, uint8_t b3){
  PORTB |= 0b00000001; // Select brightness shift (BCSEL high)
  delay(1);
  // Most significative bit is the auto-repeat function
  SPI.beginTransaction(SPISettings(14000000, MSBFIRST, SPI_MODE3));
  TLCtransfer((b3 | 0x80) & 0xFF); // Last TLC , right digit
  TLCtransfer((b2 | 0x80) & 0xFF);
  TLCtransfer((b1 | 0x80) & 0xFF); // First TLC
  SPI.endTransaction();
  writeLatch();
}






